/* loaddog.c 0.94 beta (c) 1995 Julian Assange (proff@suburbia.net),
 * All Rights Reserved.
 *
 * Loaddog will monitor various system capabilities and reboot the system in
 * as safe as possible manner if it finds that the system has fallen into
 * a state of insanity.
 *
 * Currently Loaddog monitors the following vitals and shuts down the system
 * if they repeatably fail:
 *
 *	system can create new processes
 *	system can execute new binary images
 *	system can create new file descriptors
 *	system can create new tcp sockets
 *	system load average over 15 minutes is within reason
 *	system is scheduling correctly
 *	system can allocate new virtual memory
 *	system open and read files in the file system
 *	system open and write files in the file system
 *	system file system io is functioning without curruption
 *
 * If you are running a linux1.3.51+ linux kernel, then Loaddog also supports
 * Alan Cox's /dev/watchdog (major 10, minor 130) software deadman's switch.
 *
 * Indeed, Loaddog used to be called "deadman" but collected it's current
 * title from the popular 1800's Australian short story "The Loaded Dog" by
 * Henry Lawson. In Lawson's work there is a tradition among the town's miners of
 * making bombs of various ingenious forms out of left-over mining dynamite to
 * lob into the local watering hole, much to the stunned suprise of the fish
 * who end up floating passivly to the surface and the waiting arms of
 * the miners.
 *
 * These small fish bombs are pretty interesting looking constructs and
 * it isn't too long until one of the town's mongrel curs with the attitude
 * of a golden retriever goes and grabs a live one with a long fuse and
 * proceeds to run about and "return" it. The dog fails to understand
 * why it has suddenly turned into a social lepper and no-one wants to play fetch.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/wait.h>
#include <setjmp.h>
#include <string.h>
#include <sys/param.h> /* BSDI needs this before sys/mount.h */
#include <sys/mount.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>
#include <utmp.h>
#include <sys/stat.h>
#include <errno.h>
#ifndef linux
#  include <sys/reboot.h>
#endif

#include "VERSION.h"
#include "SERIAL.h"

#ifdef linux
#  define PROC_LOADAVG
#endif

#ifdef _PATH_UTMP
#  define PATH_UTMP _PATH_UTMP
#endif
#ifndef PATH_UTMP
#  define PATH_UTMP "/etc/utmp"
#endif
#ifdef USER_PROCESS
#  define UT_ONLINE(x) ((x).ut_type==USER_PROCESS)
#else
#  define UT_ONLINE(x) ((x).ut_time>0)
#endif

#ifndef PID_FILE
#  define PID_FILE "/etc/loaddog.pid"
#endif

/* sleep(0) uses SIGALRM but usleep uses select() */
#define SLEEP(x) (usleep((x)*1000000))

jmp_buf jbuf;

struct mtable
{
	struct mtable *next;
	struct mtable *prev;
	char *name;
	int errno;
};
struct mtable *mt_head=NULL, *mt_tail=NULL;
char mt_root[64]="";

struct schain
{
	struct schain *next;
	char *str;
	int fails;
	int errno;
	char write;
};

char running=0;
char *problem="Problem? What problem?";

int dogfd;
int fdloadavg=-1;

char *argv0;

struct schain filesys_base, *filesys=&filesys_base;
struct schain wfilesys_base, *wfilesys=&wfilesys_base;
float a_loadavg=6.0;
int a_retry_read_loadavg=2;
int a_retry_loadavg=3;
int a_freq_loadavg=5;
int a_memory=1024*2048; /*2 Mb */
int a_retry_memory=3;
int a_freq_memory=3;
char *a_fork_exec="/bin/date";
int a_retry_fork_exec=2;
int a_freq_fork_exec=3;
int a_retry_sockets=4;
int a_freq_sockets=2;
int a_retry_filesys=1;
int a_freq_filesys=2;
unsigned char f_daemon=0;

void
reboot_nasty(int i)
{
#ifdef DEBUG
	fprintf(stderr, "reboot_nasty()\n");
#else
#  ifdef linux
	reboot(0xfee1dead, 672274793, 0x1234567);
#  else
	reboot(RB_AUTOBOOT);
#  endif
#endif

}
void
pexit(char *s)
{
	if (running)
		reboot_nasty(0);
	syslog(LOG_ALERT, "%s: %m", s);
	if (!f_daemon)
		perror(s);
	exit(1);
}

void *
xcalloc(int n)
{
	void *p;
	if (!(p=calloc(n, 1)))
		pexit("calloc");
	return p;
}

char *
xstrdup(char *s)
{
	char *p;
	if (!(p=malloc(strlen(s)+1)))
		pexit("malloc");
	strcpy(p, s);
	return p;
}

#ifdef DEV_WATCHDOG
void
tick()
{
	write(dogfd, "\0", 1);
	return;
}
#else
#  define tick()
#endif


void
alrmhand(int i)
{
	signal(SIGALRM, alrmhand);
	longjmp(jbuf, 1);
}


struct schain *
sadd(struct schain **sc, char *s)
{
        struct schain *p=*sc;
        p->str=s;
        p->next=xcalloc(sizeof (*p));
        p=p->next;
        *sc=p;
        return p;
}

int
mt_load()
{
	FILE *fh;
	static int before;
	struct mtable *mt=mt_head;
	char mnt[64]="", mntpoint[256]="", fstype[32]="", opts[32]="";
	if (!(fh=fopen("/etc/mtab", "r")))
		return 1;
	tick();
	for (; fscanf(fh, "%63s %255s 31%s 31%s%*[^\n]", mnt, mntpoint, fstype, opts)>0; *mnt=*mntpoint=*fstype=*opts='\0')
	{
		tick();
		if (before)
		{
			/*
			 * we are being called during terminal phase, so free birth phrase mtab.
			 * we have this in the loop as if the first fscanf didn't die the rest
			 * will probably make it.
			 */

			 while (mt)
			 {
			 	struct mtable *p;
			 	free(mt->name);
				p=mt;
				mt=mt->next;
				free(p);
			}
			before=0;
			mt=mt_head=mt_tail=NULL;
		}

		/*
		 * we only bother with physical devices and rw mounts. the system may be
		 * too wacko to umount nfs/afs et al
		 */

		if (!*mnt || !*mntpoint || strncmp(fstype, "/dev", 4) || !strncmp(opts, "ro", 2))
			continue;
		if (!mt)
			mt=mt_head=xcalloc(sizeof(*mt));
		else
		{
			mt->next=xcalloc(sizeof(*mt));
			mt->next->prev=mt;
			mt=mt->next;
		}
		mt->name=xstrdup(mnt);
	}
	if (mt)
	{
		mt->next=NULL;
		mt_tail=mt;
	}
	fclose(fh);
	tick();
	before++;
	return 0;
}

void
mt_umount()
{
	struct mtable *mt;
	/*
	 * syslog might keep a mount open
	 */
#ifndef DONT_CLOSE_SYSLOG
	closelog();
#endif
	close(0);
	close(1);
	close(2);
	sync();
	tick();
	/*
	 * we have to unmount in the reverse order to which we mounted
	 */
	for (mt=mt_tail; mt; mt=mt->prev)
	{
		if (
#ifdef MNT_FORCE
			unmount(mt->name, MNT_FORCE)
#else
			umount(mt->name)
#endif
		     			== -1)
			mt->errno=errno;
		tick();
	}
#ifdef DONT_CLOSE_SYSLOG
	/*
	 * we don't write errors in the loop above incase syslog() hangs
	 * half way through our series of unmounts
	 */
	for (mt=mt_tail; mt; mt=mt->prev)
		if (mt->errno)
		{
			errno=mt->errno;
			syslog(LOG_EMERG, "difficulty unmounting %s: %m", mt->name);
			tick();
		}
	closelog();
#endif
}

void
die(char *reason)
{
	int saved_errno=errno;
	signal(SIGALRM, alrmhand);
	if (reason)
	{
		alarm(120);
		tick();
		if (!setjmp(jbuf))
		{
			int fd;
			errno=saved_errno;
			syslog(LOG_EMERG, "%s: %m", reason);
			if ((fd=open(PATH_UTMP, O_RDONLY))!=-1)
			{
				struct utmp ut;
				char msg[512];
				char hostname[128]="";
				char tty[64]="/dev/";
				gethostname(hostname, sizeof(hostname));
				if (!*hostname)
					strcpy(hostname, "localhost");
				sprintf(msg, "\n\r\
Broadcast Message from loaddog@%s\n\r\
LOADDOG: System is rebooting in ONE MINUTE!\n\r\
LOADDOG: %s\n\r", hostname, reason);
				if (saved_errno)
					sprintf(msg+strlen(msg), "\
LOADDOG: System error %d: %s\n\r",  saved_errno, sys_errlist[saved_errno]);
				while (read(fd, &ut, sizeof(ut))==sizeof(ut))
				{
					if (UT_ONLINE(ut))
					{
						struct stat st;
						bzero(tty+5, sizeof(tty)-5);
						strncpy(tty+5, ut.ut_line, sizeof(ut.ut_line));
						if (stat(tty, &st)!=-1)
						{
							if (st.st_mode&020)
							{
								int tty_fd;
								if ((tty_fd=open(tty, O_WRONLY|O_NONBLOCK|O_NOCTTY))!=-1)
								{
									int n=0, cnt=0, blen=strlen(msg);
									char *p;
									/*
									 * spend at the most a second per user getting the
									 * warning out
									 */
									for (p=msg; blen>0 && cnt<10; cnt++)
									{
										if ((n=write(tty_fd, p, blen))<0 && errno!=EAGAIN)
											break;
										if (n<0)
											n=0;
										p+=n;
										blen-=n;
										usleep(100000);
									}
									close(tty_fd);
								}
							}
						}
					}
				}
				close(fd);
				sleep(20);
				tick();
				sleep(20);
				tick();
				sleep(20);
			}
		}
	} else
	{
		tick();
		alarm(20);
		if (!setjmp(jbuf))
			syslog(LOG_EMERG, "System going down (no reason given)!");
		tick();
	}
	signal(SIGALRM, alrmhand);
	alarm(3*60);
	if (!setjmp(jbuf))
	{
#ifdef DEBUG
		fprintf(stderr, "kill(-1, SIGTERM)\n");
#else
		kill(-1, SIGTERM);
#endif
		tick();
		SLEEP(15);
	}
	tick();
	if (!setjmp(jbuf))
	{
#ifdef DEBUG
		fprintf(stderr, "kill(-1, SIGKILL)\n");
#else
		kill(-1, SIGKILL);
#endif
	}
	tick();
	alarm(20);
	{
		unsigned char bad_mt_load=0;
		if (!setjmp(jbuf))
		{
			bad_mt_load=mt_load();
		} else
			bad_mt_load=1;
		if (bad_mt_load)
		{
			tick();
			alarm(35);
			if (!setjmp(jbuf))
				syslog(LOG_EMERG, "couldn't reload /etc/mtab; using birth time mtab");
			tick();
		}
	}
	alarm(35);
	if (!setjmp(jbuf))
		mt_umount();
	tick();
	signal(SIGALRM, reboot_nasty);
	alarm(120);
	sync();
	tick();
#ifdef DEBUG
	fprintf(stderr, "die()...reboot()\n");
#else
#  ifdef linux
	reboot(0xfee1dead, 672274793, 0x1234567);
#  else
	reboot(RB_AUTOBOOT);
#  endif
#endif
	exit(1);
}

void
segvhand(int i)
{
	signal(SIGSEGV, reboot_nasty);
	die("loaddog segmentation violation (white album)?!");
}

void
usr1hand(int i)
{
	signal(SIGUSR1, usr1hand);
	die("Forced reboot");
}

void
ok_socket()
{
	static int fails, cnt;
	int fd;
	problem="Hung in socket() check";
	if (!a_freq_sockets || (!fails && (cnt++%a_freq_sockets)))
		return;
	if ((fd=socket(PF_INET, SOCK_STREAM, 0))==-1)
	{
		if (++fails>a_retry_sockets)
			die("System unable to create new sockets!");
	} else
	{
		close(fd);
		fails=0;
	}
	return;
}


void
ok_fork_exec()
{
	static int fails, cnt;
	pid_t pid;
	int ws;
	problem="Hung in fork()/exec() check";
	if (!a_freq_fork_exec || (!fails && (cnt++%a_freq_fork_exec)))
		return;
	if (!(pid=fork()))
	{
		closelog();
#ifdef DEV_WATCHDOG
		close(dogfd);
#endif
		if (dup2(open("/dev/null", O_RDONLY), 0)==-1)
			pexit("/dev/null");
		if (dup2(open("/dev/null", O_WRONLY), 1)==-1)
			pexit("/dev/null");
		dup2(1, 2);
		execl(a_fork_exec, a_fork_exec, 0);
		_exit(1);
	}
	if (pid<0)
		goto fail;
	if (waitpid(pid, &ws, 0)!=pid)
		goto fail;
	if (!WIFEXITED(ws))
		goto fail;
	if (WEXITSTATUS(ws)!=1)
	{
		fails=0;
		return;
	}
fail:
	if (++fails>a_retry_fork_exec)
		die("System unable to fork()...exec()!");
	return;
}

#ifdef PROC_LOADAVG
void
ok_loadavg()
{
	static int fails, fail_read_loadavg, cnt;
	char line[80];
	float f1, f2, f3;
	int cc;
	problem="Hung in load average check";
	if (!a_freq_loadavg || (!fails && !fail_read_loadavg && (cnt++%a_freq_loadavg)))
		return;
	if (fdloadavg==-1)
	{
		if ((fdloadavg=open("/proc/loadavg", O_RDONLY))==-1)
		{
			fail_read_loadavg++;
			return;
		}
	} else
       		lseek(fdloadavg, 0, SEEK_SET);
	if ((cc=read(fdloadavg, line, sizeof(line)))<1)
	{
		int t=errno;
		if (++fail_read_loadavg>a_retry_filesys)
		{
			char buf[80];
			sprintf(buf, "System couldn't read the load average (failed %d times)!", fail_read_loadavg);
			errno=t;
			die(buf);
		}
		close(fdloadavg);
		fdloadavg=-1;
		return;
	} else
		fail_read_loadavg=0;
	line[cc]='\0';
	sscanf(line, "%f %f %f", &f1, &f2, &f3);
	if (f3>a_loadavg)
	{
		if (++fails>a_retry_loadavg)
		{
			char buf[80];
			sprintf(buf, "Load average = %.1f %.1f %.1f (too high)!", f1, f2, f3);
			die(buf);
		}
	} else
		fails=0;
}
#endif

void
ok_memory()
{
	static int fails, cnt;
	char *p;
	problem="Hung in memory check";
	if (!a_freq_memory || (!fails && (cnt++%a_freq_memory)))
		return;
	if (!(p=(char *)sbrk(a_memory)) || p==(char *)-1)
	{
		if (++fails>a_retry_memory)
		{
			char buf[80];
			sprintf(buf, "System couldn't sbrk() %d bytes (failed %d times)!", a_memory, fails);
			die(buf);
		}
	} else
	{
		brk(p);
		fails=0;
	}
}

void
ok_filesys()
{
	static int fails, cnt;
	struct schain *sc=&filesys_base, *high=sc;
	problem="Hung in file system check";
	if (!sc->str || !a_freq_filesys || (!fails && (cnt++%a_freq_filesys)))
		return;
	for (; sc->str; sc=sc->next)
	{
		int fd;
		tick();
		if (sc->write)
		{
			unlink(sc->str);
			fd=open(sc->str, O_CREAT|O_NOCTTY|O_RDWR|O_EXCL, 0600);
		}
		else
			fd=open(sc->str, O_RDONLY);
		if (fd==-1)
		{
			sc->errno=errno;
			sc->fails++;
		} else
		{
			char buf[400], buf2[sizeof(buf)]; /* just under a small sector */
			tick();
			if (sc->write)
			{
				memset(buf2, 0xff, sizeof(buf2));
				if (write(fd, buf2, sizeof(buf2))!=sizeof(buf2))
				{
					sc->errno=errno;
					sc->fails++;
					close(fd);
					goto skipread;
				}
				lseek(fd, 0, SEEK_SET);
				bzero(buf, sizeof(buf));
			}
			tick();
			if (read(fd, buf, sizeof(buf))<1)
			{
				sc->errno=errno;
				sc->fails++;
			}
			else
				if (sc->write && memcmp(buf, buf2, sizeof(buf)))
					sc->fails++;
			close(fd);
		}
skipread:
		tick();
		if (sc->fails > high->fails)
			high=sc;
	}
	if (high->fails)
		fails++;
	else
		fails=0;
	if (high->fails > a_retry_filesys)
	{
		char *buf=xcalloc(strlen(high->str)+80);
		sprintf(buf, "File-system dysfunctional (\"%s\" unreadable)!", high->str);
		errno=high->errno;
		die(buf);
	}
}

void
usage()
{
	fprintf(stderr, "usage: %s (%s - %s) [-d] [-OPTIONS] (look ye to the man page)\n", argv0, VERSION, SERIAL);
	exit(1);
}

char opt_c;

unsigned int
xatoi(char *s)
{
	unsigned int i;
	if (sscanf(s, "%u", &i)!=1)
	{
		fprintf(stderr, "%s error: invalid -%c value: %s\n", argv0, opt_c, s);
		usage();
	}
	return i;
}

void
xatoi2(char *s, int *i, int *k)
{
	char *p;
	if ((p=strchr(s, ',')))
	{
		*(p++)='\0';
		*k=xatoi(p);
	}
	if (*s)
		*i=xatoi(s);
}

void
main(int argc, char **argv)
{
	int c;
	argv0=argv[0];
	/*
	 * we do this incase the mm system breaks down later on.
         * at least we have grown our data hunk enough so that library
	 * routines don't have have to malloc()->sbrk()
	 */
	free(malloc(256*1024));
	while ((c=opt_c=getopt(argc, argv, "dl:L:m:M:e:E:f:F:S:vw:h"))!=-1)
		switch(c)
		{
		case 'l':
			if (sscanf(optarg, "%f", &a_loadavg)!=1)
			{
				fprintf(stderr, "%s invalid argument to -l: %s", argv0, optarg);
				usage();
			}
			break;
		case 'L':
			xatoi2(optarg, &a_freq_loadavg, &a_retry_loadavg);
			break;
		case 'm':
			a_memory=xatoi(optarg);
			break;
		case 'M':
			xatoi2(optarg, &a_freq_memory, &a_retry_memory);
			break;
		case 'e':
			a_fork_exec=optarg;
			break;
		case 'E':
			xatoi2(optarg, &a_freq_fork_exec, &a_retry_fork_exec);
			break;
		case 'f':
			sadd(&filesys, optarg);
			break;
		case 'F':
			xatoi2(optarg, &a_freq_filesys, &a_retry_filesys);
			break;
		case 'S':
			xatoi2(optarg, &a_freq_sockets, &a_retry_sockets);
			break;
		case 'd':
			f_daemon=1;
			break;
		case 'w':
			filesys->write=1;
			sadd(&filesys, optarg);
			break;
		case 'v':
		case 'h':
		default:
			usage();
		}
	chdir("/");
#ifdef PROC_LOADAVG
	if (!(fdloadavg=open("/proc/loadavg", O_RDONLY)))
		pexit("/proc/loadavg");
#endif
	mt_load();
	if (f_daemon)
	{
		int n=0;
		while(n<256) /* hey, its portable */
			close(n++);
		setsid();
		if (fork())
			_exit(0);
	}
	/* we use NDELAY just incase we can't allocate 
         * file descriptors when something nasty
	 * happens
         */
	openlog("loaddog", LOG_NDELAY|LOG_PID, LOG_KERN);
#ifndef DEBUG
#  ifdef DEV_WATCHDOG
        if ((dogfd=open("/dev/watchdog", O_WRONLY))==-1)
		pexit("/dev/watchdog");
#  endif
#else
#  ifdef DEV_WATCHDOG
        if ((dogfd=open("/dev/null", O_WRONLY))==-1)
		pexit("/dev/null");
#  endif
#endif
	nice(-10);
	tick();
	syslog(LOG_INFO, "loaddog%s daemon starting up...", VERSION);
	signal(SIGTERM, SIG_IGN);
	signal(SIGHUP, SIG_IGN);
	signal(SIGXCPU, SIG_IGN);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGCHLD, SIG_IGN);
	if (f_daemon);
		signal(SIGINT, SIG_IGN);
	signal(SIGUSR1, usr1hand);
	signal(SIGALRM, alrmhand);
	alarm(180);
	running=1;
	if (!setjmp(jbuf))
	{
		for (;;)
		{
			alarm(120);
			tick();
			ok_memory();
			tick();
#ifdef PROC_LOADAVG
			ok_loadavg();
			tick();
#endif
			ok_fork_exec();
			tick();
			ok_filesys();
			tick();
			ok_socket();
			tick();
#ifdef DEBUG
			SLEEP(1);
#else
			SLEEP(20);
			tick();
			SLEEP(20);
			tick();
			SLEEP(20);
#endif
		}
	}
	die("Hey, who stole that pesky CPU out from under me?!");
}
